from fixieai.agents.api import AgentQuery
from fixieai.agents.api import AgentResponse
from fixieai.agents.api import Embed
from fixieai.agents.api import Message
from fixieai.agents.code_shot import CodeShotAgent
from fixieai.agents.corpora import DocumentCorpus
from fixieai.agents.corpora import DocumentLoader
from fixieai.agents.oauth import OAuthHandler
from fixieai.agents.oauth import OAuthParams
from fixieai.agents.user_storage import UserStorage

__all__ = [
    "AgentQuery",
    "AgentResponse",
    "CodeShotAgent",
    "DocumentCorpus",
    "DocumentLoader",
    "Embed",
    "Message",
    "OAuthHandler",
    "OAuthParams",
    "UserStorage",
]



from fixieai.agents import AgentQuery
from fixieai.agents import AgentResponse
from fixieai.agents import CodeShotAgent
from fixieai.agents import Embed
from fixieai.agents import Message
from fixieai.agents import OAuthHandler
from fixieai.agents import OAuthParams
from fixieai.agents import UserStorage
from fixieai.client import FixieClient
from fixieai.client import get_agents
from fixieai.client import get_client
from fixieai.client import get_embeds
from fixieai.client import query

__all__ = [
    "AgentQuery",
    "AgentResponse",
    "Embed",
    "Message",
    "CodeShotAgent",
    "OAuthParams",
    "OAuthHandler",
    "UserStorage",
    "FixieClient",
    "get_agents",
    "get_client",
    "get_embeds",
    "query",
]



import dataclasses

import fastapi
import pytest
import yaml
from fastapi import testclient

import fixieai
from fixieai import agents
from fixieai.agents import code_shot

agent_id = "dummy"
BASE_PROMPT = "I am a simple dummy agent."
FEW_SHOTS = """
Q: Sample query 1
Ask Func[simple1]: Simple argument
Func[simple1] says: Simple response
A: Simple final response

Q: Sample query 2
Ask Func[simple2]: Simple argument
Func[simple2] says: Simple response
A: Simple final response
"""
CORPORA = [
    agents.DocumentCorpus(
        urls=["http://example.com/doc1.txt"], loader=agents.DocumentLoader("text")
    ),
]


@pytest.fixture(autouse=True)
def mock_token_verifier(mocker):
    return mocker.patch.object(
        code_shot._VerifiedTokenClaims,
        "from_token",
        return_value=code_shot._VerifiedTokenClaims(agent_id="fake agent id"),
    )


@pytest.fixture
def dummy_agent():
    agent = agents.CodeShotAgent(
        BASE_PROMPT,
        FEW_SHOTS,
        CORPORA,
        oauth_params=fixieai.OAuthParams(
            client_id="dummy",
            client_secret="dummy",
            auth_uri="dummy",
            token_uri="dummy",
            scopes=["dummy"],
        ),
    )

    @agent.register_func
    def simple1(query: agents.Message) -> str:
        return "Simple response 1"

    @agent.register_func()
    def simple2(query: agents.Message):
        return "Simple response 2"

    @agent.register_func(func_name="custom")
    def simple3(query):
        return "Simple response custom"

    return agent


def test_simple_agent_handshake(dummy_agent):
    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())
    client = testclient.TestClient(fast_api)
    response = client.get("/")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/yaml"
    yaml_content = yaml.load(response.content, Loader=yaml.Loader)
    assert yaml_content == {
        "base_prompt": dummy_agent.base_prompt,
        "few_shots": dummy_agent.few_shots,
        "corpora": [dataclasses.asdict(c) for c in dummy_agent.corpora],
    }


def test_simple_agent_func_calls(dummy_agent, mock_token_verifier):
    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())
    client = testclient.TestClient(fast_api, raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test Func[simple1]
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response 1", "embeds": {}}}
    mock_token_verifier.assert_called_once_with(
        "fixie-test-token", dummy_agent._jwks_client
    )

    # Test Func[simple2]
    response = client.post(
        "/simple2", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response 2", "embeds": {}}}

    # Test Func[custom]
    response = client.post(
        "/custom", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response custom", "embeds": {}}}

    # Test non-existing Func[] returns 404: Not Found
    response = client.post(
        "/simple3", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func[simple1] with bad arguments returns 422: Unprocessable Entity
    response = client.post(
        "/simple1", json={"message": {"ttt": "Howdy"}}, headers=headers
    )
    assert response.status_code == 422

    # Test Func[__init__] 404: Not Found
    response = client.post(
        "/__init__", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func without auth header returns 401: Unauthorized
    response = client.post("/simple1", json={"message": {"text": "Howdy"}})
    assert response.status_code == 403


def test_split_few_shots():
    few_shots_list = code_shot._split_few_shots(FEW_SHOTS)
    assert few_shots_list == [
        """
Q: Sample query 1
Ask Func[simple1]: Simple argument
Func[simple1] says: Simple response
A: Simple final response""",
        """Q: Sample query 2
Ask Func[simple2]: Simple argument
Func[simple2] says: Simple response
A: Simple final response""",
    ]


def good_duck_typed_func1(query):
    return "test"


def good_duck_typed_func2(query, user_storage):
    return "test"


def good_duck_typed_func3(user_storage, oauth_handler, query):
    return "test"


def bad_duck_typed_func1(*query):
    return "test"


def bad_duck_typed_func2(**user_storage):
    return "test"


def bad_duck_typed_func3(query, *user_storage):
    return "test"


def good_typed_func1(query: fixieai.Message) -> fixieai.AgentResponse:
    return fixieai.AgentResponse(fixieai.Message("test"))


def good_typed_func2(query: fixieai.Message) -> fixieai.Message:
    return fixieai.Message("test")


def good_typed_func3(user_storage: fixieai.UserStorage) -> str:
    return "test"


def good_typed_func4(
    user_storage: fixieai.UserStorage, oauth_handler: fixieai.OAuthHandler
) -> fixieai.Message:
    return fixieai.Message("test")


def good_semi_typed_func1(query: fixieai.Message):
    ...


def good_semi_typed_func2(query) -> str:
    return "test"


def good_semi_typed_func3(user_storage, query: fixieai.Message):
    ...


def good_semi_typed_func4(user_storage, oauth_handler) -> str:
    return "test"


def bad_typed_func1(query: str) -> str:
    return "test"


def bad_typed_func2(query: int) -> str:
    return "test"


def test_registering_good_and_bad_typed_funcs(dummy_agent):
    good_funcs = [
        good_typed_func1,
        good_typed_func2,
        good_typed_func3,
        good_typed_func4,
        good_duck_typed_func1,
        good_duck_typed_func2,
        good_duck_typed_func3,
        good_semi_typed_func1,
        good_semi_typed_func2,
        good_semi_typed_func3,
        good_semi_typed_func4,
    ]
    bad_funcs = [
        bad_typed_func1,
        bad_typed_func2,
        bad_duck_typed_func1,
        bad_duck_typed_func2,
        bad_duck_typed_func3,
    ]
    for good_func in good_funcs:
        dummy_agent.register_func(good_func)
        assert dummy_agent._funcs[good_func.__name__] == good_func

    for bad_func in bad_funcs:
        with pytest.raises(TypeError):
            dummy_agent.register_func(bad_func)


def test_registering_func_with_custom_name(dummy_agent):
    dummy_agent.register_func(good_typed_func1, func_name="name1")
    dummy_agent.register_func(good_typed_func2, func_name="name2")
    dummy_agent.register_func(good_typed_func3, func_name="name3")
    assert dummy_agent._funcs["name1"] == good_typed_func1
    assert dummy_agent._funcs["name2"] == good_typed_func2
    assert dummy_agent._funcs["name3"] == good_typed_func3


def test_invalid_token(dummy_agent, mock_token_verifier):
    mock_token_verifier.return_value = None

    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())
    client = testclient.TestClient(fast_api, raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 403



import dataclasses
import datetime
import json
import logging
import secrets
from typing import Dict, List, Optional
from urllib import parse

import dataclasses_json
import requests

from fixieai import constants
from fixieai.agents import api
from fixieai.agents import user_storage


@dataclasses.dataclass
class OAuthParams:
    """Encapsulates OAuth parameters, including secret, auth uri, and the scope.

    Agents who want to use OAuth flow, should declare their secrets via this object.
    """

    client_id: str
    client_secret: str
    auth_uri: str
    token_uri: str
    scopes: List[str]

    @classmethod
    def from_client_secrets_file(
        cls, secrets_path: str, scopes: List[str]
    ) -> "OAuthParams":
        """Initializes OAuth from a secrets file, e.g., as obtained from the Google
        Cloud Console.

        Args:
            secrets_path: Path to a json file holding secret values.
            scopes: A list of scopes that access needs to be requested for.
        """
        with open(secrets_path, "r") as file:
            data = json.load(file)
            secrets = data["web"] or data["installed"]
            return cls(
                secrets["client_id"],
                secrets["client_secret"],
                secrets["auth_uri"],
                secrets["token_uri"],
                scopes,
            )


class OAuthHandler:
    """
    OAuthHandler that wraps around a (OAuthParams, query) to authenticate.

    This client object provides 3 main method:
        * credentials: Returns current user's OAuth access token, or None if they are
            not authenticated.
        * get_authorization_url: Returns a url that the users can click to
            authenticate themselves.
        * authorize: Exchanges a received access code (from auth redirect callback) for
            an access token. If successful, user's storage is updated.
    """

    # OAuth keys reserved in UserStorage
    OAUTH_STATE_KEY = "_oauth_state"
    OAUTH_TOKEN_KEY = "_oauth_token"

    def __init__(
        self,
        oauth_params: OAuthParams,
        query: api.AgentQuery,
        agent_id: str,
    ):
        self._storage = user_storage.UserStorage(query, agent_id)
        self._oauth_params = oauth_params
        self._agent_id = agent_id

    def get_authorization_url(self) -> str:
        """Returns a URL to launch the authorization flow."""
        auth_state = f"{self._agent_id}:{secrets.token_urlsafe()}"
        data = {
            "response_type": "code",
            "access_type": "offline",
            "client_id": self._oauth_params.client_id,
            "scope": " ".join(self._oauth_params.scopes),
            "state": auth_state,
            "redirect_uri": constants.FIXIE_OAUTH_REDIRECT_URL,
        }
        url = self._oauth_params.auth_uri + "?" + parse.urlencode(data)
        # Store auth_state in UserStorage for validation later.
        self._storage[self.OAUTH_STATE_KEY] = auth_state
        return url

    def user_token(self) -> Optional[str]:
        """Returns current user's OAuth credentials, or None if not authorized."""
        try:
            creds_json = self._storage[self.OAUTH_TOKEN_KEY]
        except KeyError:
            return None

        if not isinstance(creds_json, str):
            logging.warning(
                f"Value at user_storage[{self.OAUTH_TOKEN_KEY!r}] is "
                f"not an OAuthCredentials json: {creds_json!r}"
            )
            del self._storage[self.OAUTH_TOKEN_KEY]
            return None

        try:
            creds = _OAuthCredentials.from_json(creds_json)
        except (TypeError, LookupError, ValueError):
            logging.warning(
                f"Value at user_storage[{self.OAUTH_TOKEN_KEY!r}] is "
                f"not a valid _OAuthCredentials json: {creds_json!r}"
            )
            del self._storage[self.OAUTH_TOKEN_KEY]
            return None

        if creds.expired:
            logging.debug(f"Credentials expired at {creds.expiry}")
            if not creds.refresh_token:
                logging.warning("No refresh token available")
                return None
            logging.debug("Refreshing credentials...")
            creds.refresh(self._oauth_params)
            self._save_credentials(creds)  # Save refreshed token to UserStorage
        return creds.access_token

    def authorize(self, state: str, code: str):
        """Authorize the received access `code` against the client secret.

        If successful, the credentials will be saved in user storage.
        """
        expected_state = self._storage[self.OAUTH_STATE_KEY]
        if state != expected_state:
            logging.warning(
                f"Unknown state token, expected: {expected_state!r} actual: {state!r}"
            )
            raise ValueError(f"Unknown state token")

        data = {
            "grant_type": "authorization_code",
            "client_id": self._oauth_params.client_id,
            "client_secret": self._oauth_params.client_secret,
            "code": code,
            "redirect_uri": constants.FIXIE_OAUTH_REDIRECT_URL,
        }
        response = _send_authorize_request(self._oauth_params.token_uri, data)
        logging.debug(
            f"OAuth auth request succeeded, lifetime={response.expires_in} "
            f"refreshable={response.refresh_token is not None}"
        )
        credentials = _OAuthCredentials(
            response.access_token,
            _get_expiry(response.expires_in),
            response.refresh_token,
        )
        self._save_credentials(credentials)

    def _save_credentials(self, credentials: "_OAuthCredentials"):
        self._storage[self.OAUTH_TOKEN_KEY] = credentials.to_json()


def _encode_iso_format(value: Optional[datetime.datetime]) -> Optional[str]:
    """Encodes a datetime to a ISO 8601 string, or None."""
    if value is None:
        return None
    return value.isoformat()


def _decode_iso_format(value: Optional[str]) -> Optional[datetime.datetime]:
    """Decodes a datetime from a ISO 8601 string, or None."""
    if value is None:
        return None
    return datetime.datetime.fromisoformat(value)


@dataclasses.dataclass
class _OAuthCredentials(dataclasses_json.DataClassJsonMixin):
    """Holds OAuth credentials, their expiration, and refresh token for a user."""

    access_token: str
    expiry: Optional[datetime.datetime] = dataclasses.field(
        default=None,
        metadata=dataclasses_json.config(
            encoder=_encode_iso_format, decoder=_decode_iso_format
        ),
    )
    refresh_token: Optional[str] = None

    @property
    def expired(self) -> bool:
        """Whether the credentials have expired."""
        return self.expiry is not None and datetime.datetime.utcnow() > self.expiry

    def refresh(self, oauth_params: OAuthParams):
        """Refreshes the access token."""
        if not self.refresh_token:
            raise ValueError("Cannot refresh without a refresh token")

        data = {
            "grant_type": "refresh_token",
            "client_id": oauth_params.client_id,
            "client_secret": oauth_params.client_secret,
            "refresh_token": self.refresh_token,
        }
        response = _send_authorize_request(oauth_params.token_uri, data)
        logging.debug(
            f"OAuth refresh request succeeded, lifetime={response.expires_in}"
        )
        self.access_token = response.access_token
        self.expiry = _get_expiry(response.expires_in)


@dataclasses.dataclass
class _OAuthTokenResponse(dataclasses_json.DataClassJsonMixin):
    """Holds an OAuth response from the OAuth server."""

    access_token: str
    scope: str
    token_type: str
    expires_in: Optional[int] = None
    refresh_token: Optional[str] = None


def _send_authorize_request(uri: str, data: Dict[str, str]) -> _OAuthTokenResponse:
    """POSTs `data` to `uri` and parses the output as an OAuthTokenResponse."""
    response = requests.post(uri, data=data)
    response.raise_for_status()
    content_type = response.headers["Content-Type"].split(";")[0].strip()
    if content_type == "application/json":
        response_dict = response.json()
    elif content_type == "application/x-www-form-urlencoded":
        # urllib parses form data into a list for each key, even if there's only one value.
        response_dict = {k: v[0] for k, v in parse.parse_qs(response.text).items()}
    else:
        raise ValueError(f"Unexpected response content type: {content_type}")
    if "error" in response_dict:
        raise ValueError("OAuth auth request failed")
    return _OAuthTokenResponse.from_dict(response_dict)


def _get_expiry(expires_in_seconds: Optional[int]) -> Optional[datetime.datetime]:
    """Returns the absolute expiration datetime from a relative expires_in_seconds."""
    if expires_in_seconds is None:
        return None
    return datetime.datetime.utcnow() + datetime.timedelta(seconds=expires_in_seconds)



import random
import string

import pytest
import validators

from fixieai.cli.agent import agent_config


@pytest.mark.parametrize("seed", range(100))
def test_slugify(seed):
    random.seed(seed)

    text = "".join(random.choices(string.ascii_letters, k=40))
    slugified = agent_config._slugify(text)
    assert len(slugified) <= len(text)
    assert validators.slug(slugified)



import importlib
import os
import sys
from typing import Tuple

from fixieai import agents
from fixieai.cli.agent import agent_config


def load_agent_from_path(
    path: str,
) -> Tuple[agent_config.AgentConfig, agents.CodeShotAgent]:
    """Loads an Agent and its config from a path."""

    path = agent_config.normalize_path(path)
    config = agent_config.load_config(path)
    agent_dir = os.path.dirname(path)

    # Inject the agent directory into PYTHONPATH
    sys.path.insert(0, agent_dir)

    entry_point_parts = config.entry_point.split(":", 1)
    module_name = entry_point_parts[0]
    attr = entry_point_parts[1] if len(entry_point_parts) == 2 else "agent"

    module = importlib.import_module(module_name)
    return config, getattr(module, attr)



from __future__ import annotations

import dataclasses
import functools
import inspect
import json
import re
import threading
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import fastapi
import jwt
import requests
import uvicorn
import yaml
from pydantic import dataclasses as pydantic_dataclasses

from fixieai import constants
from fixieai.agents import api
from fixieai.agents import corpora
from fixieai.agents import oauth
from fixieai.agents import user_storage
from fixieai.agents import utils

# Regex that controls what Func names are allowed.
ACCEPTED_FUNC_NAMES = re.compile(r"^\w+$")

# The JWT claim containing the agent ID
_AGENT_ID_JWT_CLAIM = "aid"


@pydantic_dataclasses.dataclass
class AgentMetadata:
    """Metadata for a Fixie CodeShot Agent.

    This will get sent to the Fixie platform upon handshake.
    """

    base_prompt: str
    few_shots: List[str]
    corpora: Optional[List[corpora.DocumentCorpus]] = None

    def __post_init__(self):
        utils.strip_prompt_lines(self)
        utils.validate_code_shot_agent(self)


class CodeShotAgent:
    """A CodeShot agent.

    To make a CodeShot agent, simply pass a BASE_PROMPT and FEW_SHOTS:

        BASE_PROMPT = "A summary of what this agent does; how it does it; and its
        personality"

        FEW_SHOTS = '''
        Q: <Sample query that this agent supports>
        A: <Desired response for this query>

        Q: <Another sample query>
        A: <Desired response for this query>
        '''

        agent = CodeShotAgent(BASE_PROMPT, FEW_SHOTS)

    You can have FEW_SHOTS as a single string of all your few-shots separated by 2 new
    lines, or as an explicit list of one few-shot per index.

    Your few-shots may reach out to other Agents in the fixie ecosystem by
    "Ask Agent[agent_id]: <query to pass>", or reach out to some python functions
    by "Ask Func[func_name]: <query to pass>".

    There are a series of default runtime `Func`s provided by the platform available for
    your agents to consume. For a full list of default runtime `Func`s, refer to:
        http://docs.fixie.ai/XXX

    You may also need to write your own python functions here to be consumed by your
    agent. To make a function accessible by an agent, you'd need to register it by
    `@agent.register_func`. Example:

        @agent.register_func
        def func_name(query: fixieai.Message) -> ReturnType:
            ...

        , where ReturnType is one of `str`, `fixieai.Message`, or `fixie.AgentResponse`.

    Note that in the above, we are using the decorator `@agent.register_func` to
    register this function with the agent instance we just created.

    To check out the default `Func`s that are provided in Fixie, see:
        http://docs.fixie.ai/XXX

    """

    def __init__(
        self,
        base_prompt: str,
        few_shots: Union[str, List[str]],
        corpora: Optional[List[corpora.DocumentCorpus]] = None,
        oauth_params: Optional[oauth.OAuthParams] = None,
    ):
        if isinstance(few_shots, str):
            few_shots = _split_few_shots(few_shots)

        self.base_prompt = base_prompt
        self.few_shots = few_shots
        self.corpora = corpora
        self.oauth_params = oauth_params
        self._funcs: Dict[str, Callable] = {}
        self._jwks_client = jwt.PyJWKClient(constants.FIXIE_JWKS_URL)

        if oauth_params is not None:
            # Register default Funcs.
            self.register_func(_oauth)

    def serve(
        self, agent_id: Optional[str] = None, host: str = "0.0.0.0", port: int = 8181
    ):
        """Starts serving the current agent at `{host}:{port}` via uvicorn.

        If agent_id is specified, this pings Fixie upon startup to fetch the latest prompt and fewshots.

        Args:
            agent_id: The qualified agent id (`username/handle`)
            host: The address to start listening at.
            port: The port number to start listening at.
        """
        fast_api = fastapi.FastAPI()
        fast_api.include_router(self.api_router())
        if agent_id:
            fast_api.add_event_handler(
                "startup", functools.partial(_ping_fixie_async, agent_id)
            )
        uvicorn.run(fast_api, host=host, port=port)

    def api_router(self) -> fastapi.APIRouter:
        """Returns a fastapi.APIRouter object that serves the agent."""
        router = fastapi.APIRouter()
        router.add_api_route("/", self._handshake, methods=["GET"])
        router.add_api_route("/{func_name}", self._serve_func, methods=["POST"])
        return router

    def register_func(
        self, func: Optional[Callable] = None, *, func_name: Optional[str] = None
    ) -> Callable:
        """A function decorator to register `Func`s with this agent.

        This decorator will not change the callable itself.

        Usage:

            agent = CodeShotAgent(base_prompt, few_shots)

            @agent.register_func
            def func(query):
                ...

        Optional Decorator Args:
            func_name: Optional function name to register this function by. If unset,
                the function name will be used.
        """
        if func is None:
            # Func is not passed in. It's the decorator being created.
            return functools.partial(self.register_func, func_name=func_name)

        if func_name is not None:
            if not ACCEPTED_FUNC_NAMES.fullmatch(func_name):
                raise ValueError(
                    f"Function names may only be alphanumerics, got {func_name!r}."
                )

        utils.validate_registered_pyfunc(func, self)
        name = func_name or func.__name__
        if name in self._funcs:
            raise ValueError(f"Func[{name}] is already registered with agent.")
        self._funcs[name] = func
        return func

    def _handshake(self) -> fastapi.Response:
        """Returns the agent's metadata in YAML format."""
        metadata = AgentMetadata(self.base_prompt, self.few_shots, self.corpora)
        yaml_content = yaml.dump(dataclasses.asdict(metadata))
        return fastapi.Response(yaml_content, media_type="application/yaml")

    def _serve_func(
        self,
        func_name: str,
        query: api.AgentQuery,
        credentials: fastapi.security.HTTPAuthorizationCredentials = fastapi.Depends(
            fastapi.security.HTTPBearer()
        ),
    ) -> api.AgentResponse:
        """Verifies the request is a valid request from Fixie, and dispatches it to
        the appropriate function.
        """
        token_claims = _VerifiedTokenClaims.from_token(
            credentials.credentials, self._jwks_client
        )
        if token_claims is None:
            raise fastapi.HTTPException(status_code=403, detail="Invalid token")
        elif (
            query.access_token is not None
            and query.access_token != credentials.credentials
        ):
            raise fastapi.HTTPException(status_code=403, detail="Mismatched tokens")
        else:
            query.access_token = credentials.credentials

        try:
            pyfunc = self._funcs[func_name]
        except KeyError:
            raise fastapi.HTTPException(
                status_code=404, detail=f"Func[{func_name}] doesn't exist"
            )
        kwargs = self._get_func_kwargs(
            query, token_claims, inspect.signature(pyfunc).parameters.keys()
        )
        output = pyfunc(**kwargs)
        try:
            return _wrap_with_agent_response(output)
        except TypeError:
            raise TypeError(
                f"Func[{func_name}] returned unexpected output of type {type(output)}."
            )

    def _get_func_kwargs(
        self,
        query: api.AgentQuery,
        token_claims: _VerifiedTokenClaims,
        arg_names: Iterable[str],
    ) -> Dict[str, Any]:
        kwargs: Dict[str, Any] = {}
        for arg_name in arg_names:
            if arg_name == "query":
                kwargs[arg_name] = query.message
            elif arg_name == "user_storage":
                kwargs[arg_name] = user_storage.UserStorage(
                    query, token_claims.agent_id
                )
            elif arg_name == "oauth_handler":
                assert self.oauth_params, "oauth_params is not set"
                kwargs[arg_name] = oauth.OAuthHandler(
                    self.oauth_params, query, token_claims.agent_id
                )
            else:
                raise ValueError(f"Found unknown argument {arg_name!r}.")
        return kwargs


@pydantic_dataclasses.dataclass
class _VerifiedTokenClaims:
    """Verified claims from an agent token."""

    agent_id: str

    @staticmethod
    def from_token(
        token: str, jwks_client: jwt.PyJWKClient
    ) -> Optional[_VerifiedTokenClaims]:
        try:
            public_key = jwks_client.get_signing_key_from_jwt(token)
            claims = jwt.decode(
                token,
                public_key.key,
                algorithms=["EdDSA"],
                audience=constants.FIXIE_AGENT_API_AUDIENCES,
            )
        except jwt.DecodeError:
            return None

        if _AGENT_ID_JWT_CLAIM not in claims or not isinstance(
            claims[_AGENT_ID_JWT_CLAIM], str
        ):
            # Agent id claim is required
            return None

        return _VerifiedTokenClaims(agent_id=claims[_AGENT_ID_JWT_CLAIM])


def _oauth(query: api.Message, oauth_handler: oauth.OAuthHandler) -> str:
    """Serves Func[_oauth] which is used upon auth redirect callback."""
    auth_request = json.loads(query.text)
    state = auth_request["state"]
    code = auth_request["code"]
    oauth_handler.authorize(state, code)
    return "Authorization successful!"


def _wrap_with_agent_response(
    value: Union[str, api.Message, api.AgentResponse]
) -> api.AgentResponse:
    if isinstance(value, str):
        return api.AgentResponse(api.Message(value))
    elif isinstance(value, api.Message):
        return api.AgentResponse(value)
    elif isinstance(value, api.AgentResponse):
        return value
    else:
        raise TypeError(f"Unexpected type to wrap: {type(value)}")


def _split_few_shots(few_shots: str) -> List[str]:
    """Split a long string of all few-shots into a list of few-shot strings."""
    # First, strip all lines to remove bad spaces.
    few_shots = "\n".join(line.strip() for line in few_shots.splitlines())
    # Then, split by "\n\nQ:".
    few_shot_splits = few_shots.split("\n\nQ:")
    few_shot_splits = [few_shot_splits[0]] + [
        "Q:" + few_shot for few_shot in few_shot_splits[1:]
    ]
    return few_shot_splits


def _ping_fixie_async(agent_id: str):
    """Asynchronously pings Fixie to refresh the given agent_id."""
    thread = threading.Thread(target=_ping_fixie_sync, args=(agent_id,))
    thread.start()


def _ping_fixie_sync(agent_id: str):
    response = requests.post(f"{constants.FIXIE_REFRESH_URL}/{agent_id}")
    response.raise_for_status()



import click

from fixieai.cli.session import console


@click.group(help="Session-related commands.")
def session():
    pass


@session.command("list", help="Lists sessions.")
@click.pass_context
def list_sessions(ctx):
    client = ctx.obj.client
    session_ids = client.get_sessions()
    for session_id in session_ids:
        click.secho(f"{session_id}", fg="green")


@session.command("new", help="Creates a new session and opens it.")
@click.pass_context
def new_session(ctx):
    client = ctx.obj.client
    c = console.Console(client)
    c.run()


@session.command("open", help="Opens a session.")
@click.pass_context
@click.argument("session_id")
def open_session(ctx, session_id: str):
    client = ctx.obj.client
    session = client.get_session(session_id)
    messages = session.get_messages()
    click.echo(messages)



import prompt_toolkit
import requests
import rich.console as rich_console

from fixieai import FixieClient

HISTORY_FILE = "history.txt"
textconsole = rich_console.Console()


class Console:
    """A simple console interface for Fixie."""

    def __init__(
        self,
        client: FixieClient,
        history_file: str = HISTORY_FILE,
    ):
        self._client = client
        self._session = client.create_session()
        self._history_file = history_file
        self._response_index = 0

    def run(self) -> None:
        """Run the console application."""

        textconsole.print("[blue]Welcome to Fixie!")
        textconsole.print(f"Connected to: {self._session.session_url}")
        while True:
            in_text = prompt_toolkit.prompt(
                "fixie 🚲❯ ",
                history=prompt_toolkit.history.FileHistory(self._history_file),
                auto_suggest=prompt_toolkit.auto_suggest.AutoSuggestFromHistory(),
            )
            self._query(in_text)

    def _query(self, in_text: str) -> None:
        with textconsole.status("Working...", spinner="bouncingBall"):
            try:
                self._response_index += 1
                for message in self._session.run(in_text):
                    if message["type"] != "response":
                        textconsole.print(
                            f"   [dim]@{message['sentBy']['handle']}: {message['text']}[/]"
                        )
                    else:
                        textconsole.print(f"{self._response_index}❯ {message['text']}")
            except requests.exceptions.HTTPError as e:
                textconsole.print(f"🚨 {e}")
                return



import click

from fixieai.cli.auth import oauth_flow
from fixieai.cli.auth import user_config


def validate_not_authed(ctx, param, force):
    """Click option callback for "fixie auth --force" that validates user is not
    authenticated if "--force" is not set."""
    if force:
        # --force is set
        return
    try:
        config = user_config.load_config()
        if config.fixie_api_key is not None:
            click.echo("Already authenticated. Set --force to force re-authentication.")
            ctx.exit()
    except FileNotFoundError:
        pass


@click.command("auth", help="Authorizes `fixie` to access Fixie platform.")
@click.option(
    "--force",
    is_flag=True,
    help="Forces authentication, even if the user is authenticated.",
    callback=validate_not_authed,
    expose_value=False,
)
def auth():
    fixie_api_token = oauth_flow.oauth_flow()
    try:
        config = user_config.load_config()
    except FileNotFoundError:
        config = user_config.UserConfig()

    config.fixie_api_key = fixie_api_token
    user_config.save_config(config)



import base64
import json
from typing import TYPE_CHECKING, Dict, List, MutableMapping, Union

import requests

from fixieai import constants

if TYPE_CHECKING:
    from fixieai.agents.api import AgentQuery

UserStoragePrimitives = Union[bool, int, float, str, bytes, None]
UserStorageType = Union[
    UserStoragePrimitives,
    List["UserStorageType"],
    Dict[str, "UserStorageType"],
]


class UserStorage(MutableMapping[str, UserStorageType]):
    """UserStorage provides a dict-like interface to a user-specific storage.

    Usage:
    >>> from fixieai import AgentQuery, Message
    >>> query = AgentQuery(
    ...   Message("incoming query"),
    ...   access_token="fake-access-token"
    ... )
    >>> storage = UserStorage(query, "fake-agent")
    >>> storage["key"] = "value"
    >>> storage["complex-key"] = {"key1": {"key2": [12, False, None, b"binary"]}}
    >>> assert len(storage) == 2
    >>> assert storage["complex-key"]["key1"]["key2"][-1] == b"binary"
    """

    def __init__(
        self,
        query: "AgentQuery",
        agent_id: str,
        userstorage_url: str = constants.FIXIE_USER_STORAGE_URL,
    ):
        # TODO(hessam): Remove agent_id from args once access_token includes agent_id
        #  as well.
        self._agent_id = agent_id
        self._userstorage_url = userstorage_url
        self._session = requests.Session()
        self._session.headers.update({"Authorization": f"Bearer {query.access_token}"})

    def __setitem__(self, key: str, value: UserStorageType):
        url = f"{self._userstorage_url}/{self._agent_id}/{key}"
        response = self._session.post(url, json={"data": to_json(value)})
        response.raise_for_status()

    def __getitem__(self, key: str) -> UserStorageType:
        url = f"{self._userstorage_url}/{self._agent_id}/{key}"
        try:
            response = self._session.get(url)
            response.raise_for_status()
            return from_json(response.json()["data"])
        except requests.exceptions.HTTPError as e:
            raise KeyError(f"Key {key} not found") from e

    def __contains__(self, key: object) -> bool:
        url = f"{self._userstorage_url}/{self._agent_id}/{key}"
        try:
            response = self._session.head(url)
            response.raise_for_status()
            return True
        except requests.exceptions.HTTPError as e:
            return False

    def __delitem__(self, key: str):
        url = f"{self._userstorage_url}/{self._agent_id}/{key}"
        try:
            response = self._session.delete(url)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise KeyError(f"Key {key} not found") from e

    def _get_all_keys(self):
        url = f"{self._userstorage_url}/{self._agent_id}"
        response = self._session.get(url)
        response.raise_for_status()
        return [value["key"] for value in response.json()]

    def __iter__(self):
        return iter(self._get_all_keys())

    def __len__(self):
        return len(self._get_all_keys())


JsonType = Union[None, int, float, str, bool, List["JsonType"], Dict[str, "JsonType"]]


def to_json(obj: UserStorageType) -> str:
    """Serialize a UserStorageType to a JSON string."""
    return json.dumps(to_json_type(obj))


def from_json(json_dump: str) -> UserStorageType:
    """Deserializes a UserStorageType from a JSON string."""
    return from_json_type(json.loads(json_dump))


def to_json_type(obj: UserStorageType) -> JsonType:
    """Encodes a UserStorageType to JsonType."""
    if isinstance(obj, bytes):
        return {"type": "_bytes_ascii", "data": base64.b64encode(obj).decode("ASCII")}
    elif isinstance(obj, list):
        return [to_json_type(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: to_json_type(value) for key, value in obj.items()}
    else:
        return obj


def from_json_type(obj: JsonType) -> UserStorageType:
    """Decodes a JsonType to UserStorageType."""
    if _is_bytes_encoded_json_dict(obj):
        return base64.b64decode(obj["data"])  # type: ignore
    elif isinstance(obj, list):
        return [from_json_type(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: from_json_type(value) for key, value in obj.items()}
    else:
        return obj


def _is_bytes_encoded_json_dict(obj: JsonType) -> bool:
    """Returns True if obj is a {"type": "_bytes_ascii", "data": encoded_string}."""
    if not isinstance(obj, dict):
        return False
    if set(obj.keys()) != {"type", "data"}:
        return False
    elif obj["type"] != "_bytes_ascii" or not isinstance(obj["data"], str):
        return False
    else:
        return True



import dataclasses
import os
from typing import Optional

from fixieai.cli import utils


@dataclasses.dataclass
class UserConfig(utils.DataClassYamlMixin):
    """Represents user config at ~/.config/fixie/config.yaml"""

    fixie_api_key: Optional[str] = None


def load_config(path: Optional[str] = None) -> UserConfig:
    """Loads UserConfig from the given path, or its default path.

    Args:
        path: Optional path to load config from. By default, config is loaded from
            "~/.config/fixie/config.yaml".
    """
    if path is None:
        path = os.path.expanduser("~/.config/fixie/config.yaml")
    with open(path, "r") as fp:
        return UserConfig.from_yaml(fp)


def save_config(user_config: UserConfig, path: Optional[str] = None):
    """Saves the given AgentConfig to the given path, or its default path.

    Args:
        user_config: The UserConfig object to save into file.
        path: Optional path to save. By default, it's saved to the default path at
            "~/.config/fixie/config.yaml".
    """
    if path is None:
        path = os.path.expanduser("~/.config/fixie/config.yaml")
        os.makedirs(os.path.expanduser("~/.config/fixie"), exist_ok=True)
    with open(path, "w") as fp:
        fp.write(user_config.to_yaml())



from typing import List, Optional

from pydantic import dataclasses as pydantic_dataclasses


@pydantic_dataclasses.dataclass
class DocumentLoader:
    """Document loader for a Fixie CodeShot Agent."""

    name: str


@pydantic_dataclasses.dataclass
class DocumentCorpus:
    """Document corpus for a Fixie CodeShot Agent."""

    urls: List[str]
    loader: Optional[DocumentLoader] = None



import dataclasses
import os
import re
from typing import Optional

from fixieai.cli import utils


def _current_dirname() -> str:
    """Returns current directory name."""
    return _slugify(os.path.basename(os.getcwd()))


@dataclasses.dataclass
class AgentConfig(utils.DataClassYamlMixin):
    """Represents an agent.yaml config file."""

    handle: str = dataclasses.field(default_factory=_current_dirname)
    name: str = ""
    description: str = ""
    entry_point: str = "main:agent"
    more_info_url: str = ""
    deployment_url: Optional[str] = None
    public: bool = False


def normalize_path(path: Optional[str] = None) -> str:
    """Normalizes paths to an AgentConfig.

    Args:
        path: Optional path to either a directory or YAML file. If unspecified,
            will return "{cwd}/agent.yaml".
    """
    if path is None:
        path = os.getcwd()

    if os.path.isdir(path):
        path = os.path.join(path, "agent.yaml")

    return path


def load_config(path: Optional[str] = None) -> AgentConfig:
    """Loads AgentConfig from the given path, or its default path.

    Args:
        path: Optional path to load config from. By default, config is loaded from
            "{cwd}/agent.yaml".
    """
    path = normalize_path(path)
    with open(path, "r") as fp:
        return AgentConfig.from_yaml(fp)


def save_config(agent_config: AgentConfig, path: Optional[str] = None):
    """Saves the given AgentConfig to the given path, or its default path.

    Args:
        agent_config: The AgentConfig object to save into file.
        path: Optional path to save. By default, it's saved to the default path at
            "{cwd}/agent.yaml".
    """
    path = normalize_path(path)
    with open(path, "w") as fp:
        fp.write(agent_config.to_yaml())


def _slugify(s: str) -> str:
    s = s.lower().strip()
    s = re.sub(r"[^\w\s-]", "", s)
    s = re.sub(r"[\s_-]+", "-", s)
    s = s.strip("-")
    return s



#!/usr/bin/env python3

from typing import Optional

import click

import fixieai
from fixieai import constants
from fixieai.cli.agent import commands as agent_commands
from fixieai.cli.auth import commands as auth_commands
from fixieai.cli.session import commands as session_commands


class CliContext:
    def __init__(self):
        self._client: Optional[fixieai.FixieClient] = None

    @property
    def client(self) -> fixieai.FixieClient:
        if self._client is None:
            if constants.is_authenticated():
                self._client = fixieai.FixieClient()
            else:
                click.secho(
                    "User is not authenticated. Run 'fixie auth' to authenticate, or "
                    "set the FIXIE_API_KEY environment variable, which can be obtained"
                    " from your profile page at https://app.fixie.ai/profile",
                    fg="red",
                )
                raise click.exceptions.Exit(11)
        return self._client


@click.group()
@click.pass_context
def fixie(ctx):
    """Command-line interface to the Fixie platform."""
    ctx.ensure_object(CliContext)


# Add subcommands
fixie.add_command(agent_commands.agent)
fixie.add_command(auth_commands.auth)
fixie.add_command(session_commands.session)

# Add aliases for commonly used paths
fixie.add_command(agent_commands.init_agent, "init")
fixie.add_command(session_commands.new_session, "console")


if __name__ == "__main__":
    fixie()



import requests_mock

from fixieai import agents


def test_embed_http_uri():
    with requests_mock.Mocker() as m:
        m.get("http://example.com/image.jpg", content=b"fake bytes")
        embed = agents.Embed(
            content_type="image/jpeg", uri="http://example.com/image.jpg"
        )
        assert embed.content_type == "image/jpeg"
        assert embed.uri == "http://example.com/image.jpg"
        assert embed.content == b"fake bytes"


def test_embed_data_uri():
    embed = agents.Embed(
        content_type="text/plain", uri="data:base64,SGVsbG8sIFdvcmxkIQ=="
    )
    assert embed.content_type == "text/plain"
    assert embed.content == b"Hello, World!"
    assert embed.text == "Hello, World!"

    embed.content = b"Goodbye, World!"
    assert embed.text == "Goodbye, World!"
    assert embed.uri == "data:base64,R29vZGJ5ZSwgV29ybGQh"



import secrets
import socket
import webbrowser

import click
from oauth2_client import credentials_manager

from fixieai import constants

# Fixie CLI Client ID.
CLIENT_ID = "II4FM6ToxVwSKB6DW1r114AKAuSnuZEgYehEBB-5WQA"
# Scopes to request access for.
SCOPES = ["api-access"]
# The authorization URL that users click on.
AUTHORIZE_SERVICE = f"{constants.FIXIE_API_URL}/authorize"
# The token exchange service that we use internally.
TOKEN_SERVICE = f"{constants.FIXIE_API_URL}/access/token"
# The ServiceInformation object that encapsulates all above.
SERVICE_INFORMATION = credentials_manager.ServiceInformation(
    AUTHORIZE_SERVICE,
    TOKEN_SERVICE,
    CLIENT_ID,
    None,
    SCOPES,
)


def oauth_flow() -> str:
    """Runs an interactive authorization flow with the user, and returns fixie_api_token
    if successful."""
    manager = credentials_manager.CredentialManager(SERVICE_INFORMATION)

    port = find_free_port()
    redirect_uri = f"http://localhost:{port}"
    state = secrets.token_urlsafe()
    url = manager.init_authorize_code_process(
        redirect_uri,
        state,
    )

    success = webbrowser.open(url)
    if success:
        click.echo("Your browser has been opened to visit:")
        click.echo()
        click.echo(f"    {url}")
    else:
        click.echo("Open this link on your browser to continue:")
        click.echo()
        click.echo(f"    {url}")

    # Now we wait to receive an oauth code at the redirect uri.
    code = manager.wait_and_terminate_authorize_code_process()
    # Now we exchange the code for a fixie access token.
    manager.init_with_authorize_code(redirect_uri, code)
    access_token = manager._access_token
    assert isinstance(
        access_token, str
    ), f"Invalid access token type {type(access_token)}"
    return access_token


def find_free_port() -> int:
    with socket.socket() as s:
        s.bind(("", 0))
        return int(s.getsockname()[1])



import io
import os
from contextlib import contextmanager
from typing import BinaryIO, Dict

import click
import rich.console as rich_console
import validators

import fixieai.client
from fixieai.cli.agent import agent_config
from fixieai.cli.agent import loader


@click.group(help="Agent-related commands.")
def agent():
    pass


def _validate_slug(ctx, param, value):
    while True:
        if not validators.slug(value):
            click.secho(
                f"{param.name} can be alpha numerics, underscores and dashes only."
            )
            value = click.prompt(param.prompt, default=param.default())
        else:
            return value


def _validate_url(ctx, param, value):
    while True:
        if value and not validators.url(value):
            click.secho(f"{param.name} must be a valid url.")
            value = click.prompt(param.prompt, default=param.default())
        else:
            return value


@agent.command("init", help="Creates an agent.yaml file.")
@click.option(
    "--handle",
    prompt=True,
    default=lambda: _current_config().handle,
    callback=_validate_slug,
)
@click.option(
    "--description",
    prompt=True,
    default=lambda: _current_config().description,
)
@click.option(
    "--entry-point",
    prompt=True,
    default=lambda: _current_config().entry_point,
)
@click.option(
    "--more-info-url",
    prompt=True,
    default=lambda: _current_config().more_info_url,
    callback=_validate_url,
)
@click.option(
    "--public",
    prompt=True,
    default=lambda: _current_config().public,
    type=click.BOOL,
)
def init_agent(handle, description, entry_point, more_info_url, public):
    try:
        current_config = agent_config.load_config()
    except FileNotFoundError:
        current_config = agent_config.AgentConfig()
    current_config.handle = handle
    current_config.description = description
    current_config.entry_point = entry_point
    current_config.more_info_url = more_info_url
    current_config.public = public
    agent_config.save_config(current_config)


def _current_config() -> agent_config.AgentConfig:
    """Loads current agent config, or a default if not initialized."""
    try:
        return agent_config.load_config()
    except FileNotFoundError:
        return agent_config.AgentConfig()


@agent.command("list", help="List agents.")
@click.option("--verbose", is_flag=True, help="Enable verbose output.")
@click.pass_context
def list_agents(ctx, verbose):
    client = ctx.obj.client
    agents = client.get_agents()
    for agent_id, agent in agents.items():
        click.secho(f"{agent_id}", fg="green", nl=False)
        click.echo(f": {agent['name']}")
        if verbose:
            click.echo(f"    {agent['description']}")
            if "moreInfoUrl" in agent and agent["moreInfoUrl"]:
                click.secho(f"    More info", fg="yellow", nl=False)
                click.echo(f": {agent['moreInfoUrl']}")
            click.echo()


def _validate_agent_path(ctx, param, value):
    normalized = agent_config.normalize_path(value)
    if not os.path.exists(normalized):
        raise click.BadParameter(f"{normalized} does not exist")
    return normalized


def _ensure_agent_updated(
    client: fixieai.client.FixieClient,
    config: agent_config.AgentConfig,
):
    agent = client.get_agent(config.handle)
    if agent.agent_id is None:
        agent.create_agent(
            name=config.name,
            description=config.description,
            more_info_url=config.more_info_url,
            func_url=config.deployment_url,
            published=config.public,
        )
    else:
        agent.update_agent(
            new_handle=config.handle,
            name=config.name,
            description=config.description,
            more_info_url=config.more_info_url,
            func_url=config.deployment_url,
            published=config.public,
        )

    return agent


@agent.command("serve", help="Start serving the current agent.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.option("--host", default="0.0.0.0")
@click.option("--port", type=int, default=8181)
@click.pass_context
def serve(ctx, path, host, port):
    # TODO: add tunneling and deployment url update
    config, agent_impl = loader.load_agent_from_path(path)
    agent_api = _ensure_agent_updated(ctx.obj.client, config)
    agent_impl.serve(agent_api.agent_id, host, port)


_DEPLOYMENT_BOOTSTRAP_SOURCE = """
import os
from fixieai.cli.agent import loader

if __name__ == "__main__":    
    config, agent = loader.load_agent_from_path("agent")
    agent.serve(port=int(os.getenv("PORT", "8080")))
"""


@contextmanager
def _spinner(console: rich_console.Console, text: str):
    """A context manager that renders a spinner and ✅/❌ on completion."""

    try:
        with console.status(text):
            yield
        console.print(f":white_check_mark: {text}")
    except:
        console.print(f":cross_mark: {text}")
        raise


@agent.command("deploy", help="Deploy the current agent.")
@click.argument("path", callback=_validate_agent_path, required=False)
@click.pass_context
def deploy(ctx, path):
    console = rich_console.Console()
    config = agent_config.load_config(path)
    agent_api = _ensure_agent_updated(ctx.obj.client, config)

    if config.deployment_url is None:
        # Deploy the agent to fixie with some bootstrapping code.
        file_streams: Dict[str, BinaryIO] = {}
        deploy_root = os.path.dirname(path)
        for root, dirs, files in os.walk(deploy_root, topdown=True):
            # Exclude any hidden files by removing them from the dirs list.
            # As per https://docs.python.org/3/library/os.html#os.walk
            # removing them from the dirs list with topdown=True skips
            # them from the walk.
            for i, dir in reversed(list(enumerate(dirs))):
                if dir.startswith("."):
                    console.print(
                        f"Ignoring hidden directory {os.path.join(root, dir)}",
                        style="grey53",
                    )
                    # N.B. We're iterating in reverse to avoid needing to adjust i.
                    del dirs[i]

            for filename in files:
                if filename.startswith("."):
                    console.print(
                        f"Ignoring hidden file {os.path.join(root, filename)}",
                        style="grey53",
                    )
                    continue

                full_path = os.path.join(root, filename)
                upload_path = os.path.join(
                    "agent", os.path.relpath(full_path, deploy_root)
                )
                file_streams[upload_path] = open(full_path, "rb")

        file_streams["main.py"] = io.BytesIO(_DEPLOYMENT_BOOTSTRAP_SOURCE.encode())
        with _spinner(console, "Deploying..."):
            ctx.obj.client.deploy_agent(
                config.handle,
                file_streams,
            )

    # Trigger a refresh with the updated deployment
    with _spinner(console, "Refreshing..."):
        ctx.obj.client.refresh_agent(config.handle)



from typing import TextIO, Type, TypeVar, Union

import dataclasses_json
import yaml

A = TypeVar("A", bound="DataClassYamlMixin")


class DataClassYamlMixin(dataclasses_json.DataClassJsonMixin):
    @classmethod
    def from_yaml(cls: Type[A], config: Union[str, TextIO]) -> A:
        return cls.from_dict(yaml.safe_load(config))

    def to_yaml(self) -> str:
        return yaml.dump(self.to_dict(), sort_keys=False)  # type: ignore



#!/usr/bin/env python3

from __future__ import annotations

import datetime
import threading
import time
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional

from gql import gql

if TYPE_CHECKING:
    import fixieai.client as fixie_client

    FixieClient = fixie_client.FixieClient
else:
    FixieClient = Any


class Session:
    """Represents a single session with the Fixie system.

    Args:
        client: The FixieClient instance to use.
        session_id: The ID of the session to use. If not provided, a new
            session will be created.
    """

    def __init__(
        self,
        client: FixieClient,
        session_id: Optional[str] = None,
    ):
        self._client = client
        self._gqlclient = self._client.gqlclient
        self._session_id = session_id
        if session_id:
            # Test that the session exists.
            _ = self.get_metadata()
        else:
            self._session_id = self._create_session()
        self._last_message_timestamp: Optional[datetime.datetime] = None

    @property
    def session_id(self) -> Optional[str]:
        """Return the session ID used by this Fixie client."""
        return self._session_id

    @property
    def session_url(self) -> str:
        """Return the URL of the Fixie session."""
        return f"{self._client.url}/sessions/{self.session_id}"

    def clone(self) -> "Session":
        """Return a new Session instance with the same configuration."""
        return Session(self._client.clone(), session_id=self._session_id)

    def _create_session(self) -> str:
        """Create a new session."""
        assert self._session_id is None

        query = gql(
            """
            mutation CreateSession {
                createSession(sessionData: {}) {
                    session {
                        handle
                    }
                }
            }
            """
        )
        result = self._gqlclient.execute(query)
        if "createSession" not in result or result["createSession"] is None:
            raise ValueError(f"Failed to create Session")
        assert isinstance(result["createSession"], dict)
        assert isinstance(result["createSession"]["session"], dict)
        assert isinstance(result["createSession"]["session"]["handle"], str)
        return result["createSession"]["session"]["handle"]

    def get_metadata(self) -> Dict[str, Any]:
        """Return metadata about this session."""

        query = gql(
            """
            query getSession($session_id: String!) {
                sessionByHandle(handle: $session_id) {
                    handle
                    name
                    description
                }
            }
        """
        )
        result = self._gqlclient.execute(
            query, variable_values={"session_id": self._session_id}
        )
        assert "sessionByHandle" in result and isinstance(
            result["sessionByHandle"], dict
        )
        return result["sessionByHandle"]

    def delete_session(self) -> None:
        """Delete the current session."""
        query = gql(
            """
            mutation DeleteSession($handle: String!) {
                deleteSession(handle: $handle) {
                    session {
                        handle
                    }
                }
            }
        """
        )
        _ = self._gqlclient.execute(query, variable_values={"handle": self._session_id})

    def get_embeds(self) -> List[Dict[str, Any]]:
        """Return the Embeds attached to this Session."""
        query = gql(
            """
            query getEmbeds($handle: String!) {
                sessionByHandle(handle: $handle) {
                    embeds {
                        key
                        embed {
                            id
                            contentType
                            created
                            contentHash
                            owner {
                                username
                            }
                            url
                        }
                    }
                }
            }
        """
        )
        result = self._gqlclient.execute(
            query, variable_values={"handle": self._session_id}
        )
        assert (
            "sessionByHandle" in result
            and isinstance(result["sessionByHandle"], dict)
            and isinstance(result["sessionByHandle"]["embeds"], list)
        )
        return result["sessionByHandle"]["embeds"]

    def get_messages(self) -> List[Dict[str, Any]]:
        """Return the messages that make up this session."""
        query = gql(
            """
            query getMessages($handle: String!) {
                sessionByHandle(handle: $handle) {
                    messages {
                        id
                        text
                        sentBy {
                            handle
                        }
                        type
                        inReplyTo { id }
                        timestamp
                    }
                }
            }
        """
        )
        result = self._gqlclient.execute(
            query, variable_values={"handle": self._session_id}
        )
        assert (
            "sessionByHandle" in result
            and isinstance(result["sessionByHandle"], dict)
            and isinstance(result["sessionByHandle"]["messages"], list)
        )
        return result["sessionByHandle"]["messages"]

    def add_message(self, text: str) -> str:
        """Add a message to this Session. Returns the added message text."""
        query = gql(
            """
            mutation Post($handle: String!, $text: String!) {
                sendSessionMessage(messageData: {session: $handle, text: $text}) {
                    message {
                        text
                    }
                }
            }
            """
        )
        result = self._gqlclient.execute(
            query, variable_values={"handle": self._session_id, "text": text}
        )
        assert isinstance(result["sendSessionMessage"]["message"]["text"], str)
        return result["sendSessionMessage"]["message"]["text"]

    def query(self, text: str) -> str:
        """Run a single query against the Fixie API and return the response."""
        self.add_message(text)
        # The reply to the query comes in as the most recent 'response' message in the session.
        response = self.get_messages()[-1]
        assert isinstance(response["text"], str)
        return response["text"]

    def run(self, text: str) -> Generator[Dict[str, Any], None, None]:
        """Run a query against the Fixie API, returning a generator that yields
        messages."""

        # Run the query in the background, and continue polling for replies.
        background_client = self.clone()
        threading.Thread(target=background_client.add_message, args=(text,)).start()

        response_received = False
        while not response_received:
            time.sleep(1)
            messages = self._get_messages_since(self._last_message_timestamp)
            for message in messages:
                timestamp = datetime.datetime.fromisoformat(message["timestamp"])
                self._last_message_timestamp = timestamp
                response_received = message["type"] == "response"
                yield message

    def _get_messages_since(
        self, timestamp: Optional[datetime.datetime]
    ) -> List[Dict[str, Any]]:
        """Return all messages since the given timestamp."""
        messages = self.get_messages()
        return [
            m
            for m in messages
            if timestamp is None or m["timestamp"] > timestamp.isoformat()
        ]
